package tools.descartes.prisma.core.agent;

import java.util.LinkedList;
import java.util.List;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import org.eclipse.emf.cdo.CDOObject;
import org.eclipse.emf.ecore.EObject;

import tools.descartes.prisma.core.MessageBus;
import tools.descartes.prisma.core.ModelRepository;
import tools.descartes.prisma.core.Transaction;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapter;
import tools.descartes.prisma.core.adapter.ModelSkeletonAdapterFactory;
import tools.descartes.prisma.core.adapter.ModelSkeletonDelta;
import tools.descartes.prisma.core.exceptions.CommitException;
import tools.descartes.prisma.core.owner.OwnershipManager;
import tools.descartes.prisma.core.scopes.AgentScope;
import tools.descartes.prisma.core.scopes.ModelNotification;
import tools.descartes.prisma.core.scopes.SystemGlobalScope;
import tools.descartes.prisma.model.scope.Agent;
import tools.descartes.prisma.model.skeleton.ModelSkeleton;

public class AgentController {

	private static final int MAX_RETRIES = 3;

	private Agent agent;
	private final MessageBus messageBus;
	private final ModelRepository repository;
	private final ScheduledExecutorService workerThread = Executors.newScheduledThreadPool(1);
	private final String scopePath;
	private final String id;
	private AgentScope scope;
	private OwnershipManager ownership;
	private ModelSkeletonAdapterFactory adapterFactory;
	private RemoveFilter removeFilter;

	private volatile boolean running = false;

	public AgentController(ModelRepository repository, MessageBus messageBus, String scopePath, String id) {
		this.messageBus = messageBus;
		this.repository = repository;
		this.scopePath = scopePath;
		this.id = id;
	}

	public Agent getAgent() {
		return agent;
	}

	public AgentScope getContainerScope() {
		return scope;
	}

	public OwnershipManager getOwnershipManager() {
		return ownership;
	}

	public void setRemoveFilter(RemoveFilter filter) {
		this.removeFilter = filter;
	}

	public Future<?> start() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (!running) {
					startAsync();
				}
				running = true;
			}
		});
	}

	public Future<?> stop() {
		return workerThread.submit(new Runnable() {
			public void run() {
				if (running) {
					stopAsync();
				}
				running = false;
			}
		});
	}

	public void apply(ModelSkeleton skeleton) {
		int retries = 0;

		while (retries <= MAX_RETRIES) {

			try (Transaction transaction = getModelRepository().createTransaction()) {

				scope.synchronizeModelSkeleton(this, transaction, skeleton);

				boolean success = true;

				try {
					transaction.commit();
				} catch (CommitException ex) {
					System.err.println("Commit conflict. Retry...");
					success = false;
				}

				List<ModelSkeletonAdapter> adapters = new LinkedList<>();
				for (EObject curObj : skeleton.eContents()) {
					ModelSkeletonAdapter curAdapter = adapt(curObj, ModelSkeletonAdapter.class);
					adapters.add(curAdapter);
				}
				// IMPORTANT: commit after all model skeleton adapters have been
				// synchronized and merged.
				for (ModelSkeletonAdapter curAdapter : adapters) {
					ModelSkeletonDelta delta = curAdapter.commitFinished(success);

					if (success) {
						// Send notifications for all changes
						for (EObject curObject : delta.getNewObjects()) {
							scope.sendNotifications(ModelNotification.NEW_NOTIFICATION, (CDOObject) curObject);
						}
						for (EObject curObject : delta.getRemovedObjects()) {
							scope.sendNotifications(ModelNotification.REMOVED_NOTIFICATION, (CDOObject) curObject);
						}
					}
				}

				// The commit was successful -> just return to caller, otherwise
				// retry
				if (success) {
					return;
				}
			}
			retries++;
		}

		// Maximum number of retries reached. Failure
		System.err.println("Maximum number of retries reached. Commit was not successful.");
	}

	protected ScheduledFuture<?> schedule(Runnable command, long period, TimeUnit unit) {
		return workerThread.scheduleAtFixedRate(command, 0, period, unit);
	}

	protected void startAsync() {
		messageBus.connect();
		repository.connect();
		SystemGlobalScope systemScope = getModelRepository().getSystemScope();
		systemScope.create();
		this.scope = (AgentScope) getModelRepository().getScope(scopePath);
		this.scope.create();
		this.agent = scope.addAgent(id);
		this.adapterFactory = new ModelSkeletonAdapterFactory(repository, this);
		this.ownership = new OwnershipManager(getModelRepository(), scope.getScopeConfiguration(), this.agent);

		// Deliver the notifications for the delegated objects which were
		// registered before this agent started.
		// TODO:
	}

	protected void stopAsync() {
		try {
			scope.removeAgent(this, removeFilter);
		} finally {
			repository.disconnect();
			messageBus.disconnect();
		}
	}

	public ModelRepository getModelRepository() {
		return repository;
	}

	public void handleNotification(ModelNotification notification) {
	}

	public String[] getDelegationDestinations(EObject object) {
		return new String[0];
	}

	protected void checkRunning() {
		if (!running) {
			throw new IllegalStateException("The agent is not running.");
		}
	}

	@SuppressWarnings("unchecked")
	public <T> T adapt(EObject obj, Class<T> adapterClass) {
		return (T) adapterFactory.adapt(obj, adapterClass);
	}
}
