package tools.descartes.prisma.core.impl;

import java.io.IOException;
import java.util.Map;
import java.util.concurrent.TimeoutException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import com.rabbitmq.client.AMQP;
import com.rabbitmq.client.AMQP.BasicProperties;
import com.rabbitmq.client.Channel;
import com.rabbitmq.client.Connection;
import com.rabbitmq.client.ConnectionFactory;
import com.rabbitmq.client.DefaultConsumer;
import com.rabbitmq.client.Envelope;

import tools.descartes.prisma.core.MessageBus;
import tools.descartes.prisma.core.exceptions.MessagingException;

public class RabbitMQMessageBus implements MessageBus {

	private static final int RABBITMQ_PORT = 5672;
	private static final String RABBITMQ_VHOST = "prisma";

	private final Log log = LogFactory.getLog(RabbitMQMessageBus.class);

	private final String userName;
	private final String password;
	private final String host;

	private Connection connection;
	private Channel channel;

	public RabbitMQMessageBus(String host, String userName, String password) {
		super();
		this.userName = userName;
		this.password = password;
		this.host = host;
	}

	@Override
	public void connect() {
		ConnectionFactory connectionFactory = new ConnectionFactory();
		connectionFactory.setUsername(userName);
		connectionFactory.setPassword(password);
		connectionFactory.setVirtualHost(RABBITMQ_VHOST);
		connectionFactory.setHost(host);
		connectionFactory.setPort(RABBITMQ_PORT);
		try {
			connection = connectionFactory.newConnection();
			channel = connection.createChannel();
		} catch (IOException | TimeoutException e) {
			log.error("Error connecting to RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	@Override
	public void disconnect() {
		try {
			if (channel != null) {
				channel.close();
			}
			if (connection != null) {
				connection.close();
			}
		} catch (IOException | TimeoutException e) {
			log.error("Error disconnecting from RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	public void createExchange(String exchange, ExchangeType type) {
		try {
			String typeString = "";
			switch (type) {
			case DIRECT:
				typeString = "direct";
				break;
			case FANOUT:
				typeString = "fanout";
				break;
			case TOPIC:
				typeString = "topic";
				break;
			case HEADERS:
				typeString = "headers";
				break;
			}
			channel.exchangeDeclare(exchange, typeString, false);
		} catch (IOException e) {
			log.error("Error creating exchange in RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	public void removeExchange(String exchange) {
		try {
			channel.exchangeDelete(exchange);
		} catch (IOException e) {
			log.error("Error removing exchange in RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	public void createQueue(String queue) {
		try {
			channel.queueDeclare(queue, false, false, false, null);
		} catch (IOException e) {
			log.error("Error creating queue in RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	@Override
	public void removeQueue(String queue) {
		try {
			channel.queueDelete(queue);
		} catch (IOException e) {
			log.error("Error removing queue in RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	public void bindQueue(String queue, String exchange, String routingKey) {
		try {
			channel.queueBind(queue, exchange, routingKey);
		} catch (IOException e) {
			log.error("Error creating queue in RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	public void publishMessage(String exchange, String routingKey, String message, Map<String, Object> headers) {
		byte[] bodyBytes = message.getBytes();
		try {
			channel.basicPublish(exchange, routingKey, new AMQP.BasicProperties.Builder().headers(headers).build(),
					bodyBytes);
		} catch (IOException e) {
			log.error("Error publishing message to RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

	public void addMessageListener(String queue, final Listener listener) {
		try {
			boolean autoAck = false;
			channel.basicConsume(queue, autoAck, new DefaultConsumer(channel) {
				@Override
				public void handleDelivery(String consumerTag, Envelope envelope, BasicProperties properties,
						byte[] body) throws IOException {
					listener.newMessage(envelope.getRoutingKey(), properties.getHeaders(), new String(body));
					channel.basicAck(envelope.getDeliveryTag(), false);
				}
			});
		} catch (IOException e) {
			log.error("Error registering consumer at RabbitMQ server", e);
			throw new MessagingException(e);
		}
	}

}
